module RedMatryoshka
  class Base
    attr_accessor :sub_id
    @sub_id = nil

    def set_sub_id(sub_id)
      @sub_id = sub_id
      self
    end

    def class_to_fetch_from
      @class_to_fetch_from ||= self.class.class_to_fetch_from
    end

    def classes_to_depends_on
      @classes_to_depends_on ||= self.class.cache_classes_to_depends_on || {}
    end

    def class_to_depends_on_with_dependencies
      @class_to_depends_on_with_dependencies ||= begin
        classes_to_depends_on.inject({}) do |hashes, (k, v)|
          sub_dependencies = {
            sub_dependencies: cache_klass_with_module(k).constantize.new.classes_to_depends_on
          }
          hashes.merge! k => v.merge(sub_dependencies)
        end
      end
    end

    def classes_dependencies_of
      @cache_classes_dependencies ||= self.class.cache_classes_dependencies || []
    end

    def classes_to_listen_for
      default_listen_for = { self.class.class_to_fetch_from => { sub: nil, record: :self } }
      @classes_to_listen_for ||= self.class.classes_to_listen_for || default_listen_for
    end

    def to_hash(object)
      object.attributes.symbolize_keys
    end

    def after_fetch(hashes)
      # Instantiate me if you want to edit data
      hashes
    end

    def cache_key(id, prefix = true)
      key_prefix = RedMatryoshka.configuration.key_prefix
      klass_name = self.class.name.demodulize # Only retrieve class without module name

      key_parts = klass_name.split(":")
      key_parts << @sub_id unless @sub_id.nil?
      if prefix && key_parts[0].casecmp(key_prefix) != 0
        key_parts.insert(0, key_prefix)
      end

      key_parts = key_parts.map do |part|
        part.to_s.split(/(?=[A-Z])/).map(&:downcase).join("_")
      end

      key_parts << id
      key_parts.join(":")
    end

    def cache_klass_with_module(klass)
      if RedMatryoshka.configuration.module.present?
        "#{RedMatryoshka.configuration.module}::#{klass}"
      else
        klass
      end
    end

    class << self
      DEFAULT_DEPENDS_ON_OPTIONS = { for: nil, as: :hash }.freeze
      DEFAULT_LISTEN_FOR_OPTIONS = { sub: nil, record: :self }.freeze
      attr_reader :class_to_fetch_from,
                  :cache_classes_to_depends_on,
                  :cache_classes_dependencies,
                  :classes_to_listen_for

      def fetch(*ids)
        RedMatryoshka::Cache.new(new).fetch(ids)
      end

      def fetch_sub(sub_id, ids = [])
        klass = new.set_sub_id sub_id

        RedMatryoshka::Cache.new(klass).fetch(ids)
      end

      def fetch_from(klass)
        @class_to_fetch_from = transformed_class klass
      end

      def listen_for(klass, options = {})
        @classes_to_listen_for = {} if @classes_to_listen_for.nil?

        klass_options = DEFAULT_LISTEN_FOR_OPTIONS.merge options
        @classes_to_listen_for[transformed_class(klass)] = klass_options
      end

      def depends_on(klass, options = {})
        klass_to_depends_on = transformed_class klass

        @cache_classes_to_depends_on = {} if @cache_classes_to_depends_on.nil?

        unless @cache_classes_to_depends_on.include? klass_to_depends_on
          klass_options = DEFAULT_DEPENDS_ON_OPTIONS.merge options
          @cache_classes_to_depends_on[klass_to_depends_on] = klass_options
        end
      end

      def depends_on_multiple(klass, options = {})
        for_option = options[:for] || nil
        depends_on(klass, options.merge(for: for_option, as: :array))
      end

      def dependency_of(klass)
        klass_that_depends_of = transformed_class klass

        @cache_classes_dependencies = [] if @cache_classes_dependencies.nil?
        unless @cache_classes_dependencies.include? klass_that_depends_of
          @cache_classes_dependencies << klass_that_depends_of
        end
      end

      def transformed_class(klass)
        klass.to_s.split("_").map(&:capitalize).join
      end
    end
  end
end
